from __future__ import annotations
from abc import ABC
from ..relations.relation import Relation


# předek pro binární operace relační algebry, abstraktní třída
class BinaryOperation(Relation, ABC):
    """Abstract ancestor for all binary operations on relations

    Attributes
    ----------
    left : Relation
        Left side of the operation, unmodified
    right : Relation
        Right side of the operation, unmodified
    """

    def __init__(self, schema: list, left: Relation, right: Relation):
        """
        Parameters
        ----------
        schema : list
            Schema of the resulting relation
        left : Relation
            Left side of the operation
        right : Relation
            Right side of the operation
        """
        super().__init__(schema)
        self._left = left
        self._right = right

    def get_left(self) -> Relation:
        """Gets original relation from the left side of the operation"""
        return self._left

    def get_right(self) -> Relation:
        """Gets original relation from the right side of the operation"""
        return self._right



class RelationalSchema:
    """Class providing static functions for work with relational schemas in the form of lists"""

    def cross(l_schema: list, r_schema: list) -> list:
        """Creates a new schema for cross join of relations

        Parameters
        ----------
        l_schema : list
            A list of attribute names
        r_schema : list
            Another list of attribute names

        Returns
        -------
        list
            Combined list of attribute names, duplicate attributes are renamed and kept
        """
        new_schema: list = l_schema.copy()
        for attr in r_schema:
            if attr in new_schema:
                unique_name = RelationalSchema.get_unique_attr_name(attr, new_schema)
                new_schema.append(unique_name)
            else:
                new_schema.append(attr)
        return new_schema

    def join(l_schema: list, r_schema: list) -> (list, list):
        """Creates a new schema for natural join of relations

        Parameters
        ----------
        l_schema : list
            A list of attribute names
        r_schema : list
            Another list of attribute names

        Returns
        -------
        (list, list)
            Tuple containing a joined schema and a list of common attributes
        """
        new_schema = l_schema.copy()
        common_attrs: list = []
        for attr in r_schema:
            if attr in new_schema:
                common_attrs.append(attr)
            else:
                new_schema.append(attr)
        return (new_schema, common_attrs)

    def print(schema: list):
        """Prints attribute names in a line separated by spaces

        Parameters
        ----------
        schema : list
            A list of attribute names to print
        """
        for attr in schema:
            print(attr, end=" ")
        print("")

    def get_unique_attr_name(attr_name: str, attr_names: list) -> str:
        """Creates a new attribute name not contained in a list of names by appending '

        Parameters
        ----------
        attr_name : str
            Non-unique attribute name
        attr_names : list
            List of attribute names to check uniqueness

        Returns
        -------
        str
            Unique attribute name
        """
        new_name = attr_name
        while new_name in attr_names:
            new_name += "'"
        return new_name



from .binary_operation import BinaryOperation
from ..utils.relational_schema import RelationalSchema
from ..relations.relation import Relation
from ..utils.list_operations import dict_from_lists, dict_values
from itertools import islice


class CrossJoin(BinaryOperation):
    """Cross join (Cartesian product) of two relations"""

    def __init__(self, left: Relation, right: Relation):
        """Multiplies left relation with right

        Not commutative

        Parameters
        ----------
        left : Relation
            Left side of the operation
        right : Relation
            Right side of the operation
        """
        super().__init__(
            RelationalSchema.cross(left.get_schema(), right.get_schema()),
            left,
            right,
        )

    def can_be_infinite(self):
        return self.get_left().can_be_infinite() or self.get_right().can_be_infinite()

    def can_is_member_loop(self):
        return (
            self.get_left().can_is_member_loop()
            or self.get_right().can_is_member_loop()
        )

    def is_member(self, tpl: dict) -> bool:
        l_tuple, r_tuple = self._split_tuple(tpl)
        return self.get_left().is_member(l_tuple) and self.get_right().is_member(
            r_tuple
        )

    def members(self):
        """Combines every tuple from the left relation with every tuple from the right"""
        if not self.get_left().can_be_infinite():
            for r_tuple in self.get_right().members():
                for l_tuple in self.get_left().members():
                    yield self._cross_tuple(l_tuple, r_tuple)
        elif not self.get_right().can_be_infinite():
            for l_tuple in self.get_left().members():
                for r_tuple in self.get_right().members():
                    yield self._cross_tuple(l_tuple, r_tuple)
        else:
            limit = 0
            while True:
                for i in range(limit + 1):
                    l = next(islice(self.get_left().members(), i, None))
                    r = next(islice(self.get_right().members(), limit - i, None))
                    yield self._cross_tuple(l, r)
                limit += 1

    def _cross_tuple(self, l_tuple: dict, r_tuple: dict) -> dict:
        """Helper function to combine two tuples

        Parameters
        ----------
        l_tuple : dict
            Tuple from the left side of the cross operation
        r_tuple : dict
            Tuple from the right side of the cross operation

        Returns
        -------
        dict
            A tuple with a new schema and both tuple's combined values
        """
        return dict_from_lists(
            self.get_schema(),
            list(l_tuple.values()) + list(r_tuple.values()),
        )

    def _split_tuple(self, tuple: dict) -> tuple[dict, dict]:
        """Helper function to split a combined tuple into it's original ones

        Parameters
        ----------
        tuple : dict
            Combined/crossed tuple to be split

        Returns
        -------
        tuple[dict, dict]
            A tuple of two relational tuples from the original relations
        """
        l_schema = self.get_left().get_schema()
        r_schema = self.get_right().get_schema()
        s_schema = self.get_schema()

        cross_l_schema = [attr if attr in s_schema else attr + "'" for attr in l_schema]
        cross_r_schema = [attr if attr in s_schema else attr + "'" for attr in r_schema]

        l_tuple = dict_from_lists(l_schema, dict_values(tuple, cross_l_schema))
        r_tuple = dict_from_lists(r_schema, dict_values(tuple, cross_r_schema))

        return (l_tuple, r_tuple)



"""This module contains tools for creating and operating on both finite and infinite relations

Constructor Functions
---------------------
    list_relation(schema)
    integer_relation(schema)
    general_integer_relation(schema, condition)
    add_relation(schema)

Basic Operations
----------------
    union(left, right)
    difference(left, right)
    cross_join(left, right)
    projection(relation, new_schema)
    selection(relation, condition, condition_params)

Derived Operations
------------------
    intersection(left, right)
    natural_join(left, right)
    division(left, right)
"""
from .relations.relation import Relation
from .relations.list_relation import ListRelation
from .relations.infinite_relations import (
    GeneralIntegerRelation,
    AddRelation,
    IntegerRelation,
    FibonacciRelation,
    EqualRelation,
    UnequalRelation,
    WordRelation,
)
from .operations.union import Union
from .operations.difference import Difference
from .operations.cross_join import CrossJoin
from .operations.projection import Projection
from .operations.selection import Selection
from .operations.derived import *
from .utils.conditions import *
from .utils.list_operations import *


# basic constructor functions
def list_relation(schema: list, tuples: list) -> ListRelation:
    """Creates a finite relation from a list of attribute names and a list of tuples

    Parameters
    ----------
    schema : list
        List of attribute names
    tuples : list, optional
        A list of tuples to initialize the relation, by default []

    Returns
    -------
    ListRelation
        New finite relation
    """
    return ListRelation(schema, tuples)


def integer_relation(schema: list = ["n"]) -> IntegerRelation:
    """Creates unary integer relation with a given one-attribute schema

    Parameters
    ----------
    schema : list, optional
        List of attribute names, by default ["n"]

    Returns
    -------
    IntegerRelation
        New unary relation of non-negative integers
    """
    return IntegerRelation(schema)


def general_integer_relation(
    schema: list, condition: callable
) -> GeneralIntegerRelation:
    """Creates an infinite relation of integer tuples determined by a given schema and condition

    Parameters
    ----------
    schema : list
        List of attribute names
    condition : callable
        Condition that tuples must satisfy (i.e. lambda x: x % 2 == 0)

    Returns
    -------
    GeneralIntegerRelation
        New infinite relation of integer tuples, determined by the condition
    """
    return GeneralIntegerRelation(schema, condition)


# premade infinite relations
def add_relation(schema: list = ["x", "y", "sum"]) -> AddRelation:
    """Creates an add relation with a given schema

    Parameters
    ----------
    schema : list, optional
        List of attribute names, by default ["x", "y", "sum"]

    Returns
    -------
    AddRelation
        New infinite relation of all possible integer sums
    """
    return AddRelation(schema)


def fibonacci_relation() -> FibonacciRelation:
    return FibonacciRelation()


def equal_relation(schema: list = ["x", "y"]) -> EqualRelation:
    return EqualRelation(schema)


def unequal_relation() -> UnequalRelation:
    return UnequalRelation()


def word_relation(schema: list = ["word"]) -> WordRelation:
    return WordRelation(schema)


# basic operations
def union(left: Relation, right: Relation) -> Union:
    """Unites left relation with right

    Parameters
    ----------
    left : Relation
        Left side of the operation
    right : Relation
        Right side of the operation

    Returns
    -------
    Union
        New united relation
    """
    return Union(left, right)


def difference(left: Relation, right: Relation) -> Difference:
    """Subtracts right relation from left

    Not commutative

    Parameters
    ----------
    left : Relation
        Left side of the operation
    right : Relation
        Right side of the operation

    Returns
    -------
    Difference
        New differential relation
    """
    return Difference(left, right)


def cross_join(left: Relation, right: Relation) -> CrossJoin:
    """Multiplies left relation with right

    Not commutative

    Parameters
    ----------
    left : Relation
        Left side of the operation
    right : Relation
        Right side of the operation

    Returns
    -------
    CrossJoin
        New Cartesian product of relations
    """
    return CrossJoin(left, right)


def projection(relation: Relation, new_schema: list) -> Projection:
    """Projects relation onto new schema

    Parameters
    ----------
    relation : Relation
        Input relation of the operation
    new_schema : list
        New schema to project the relation onto

    Returns
    -------
    Projection
        New projected relation
    """
    return Projection(relation, new_schema)


def selection(
    left: Relation, condition: callable, condition_param_names: list = []
) -> Selection:
    """Restricts the relation to tuples that satisfy a condition

    Parameters
    ----------
    relation : Relation
        Input relation of the operation
    condition : function
        Condition, that tuples must satisfy
    condition_param_names : list
        List of attribute names to be evaluated by the condition

    Returns
    -------
    Selection
        New restricted relation
    """
    return Selection(left, condition, condition_param_names)


# derived operations are already exported in the correct format



from itertools import zip_longest
from ..relations.relation import Relation
from .binary_operation import BinaryOperation
from ..utils.list_operations import flatten_list


class Union(BinaryOperation):
    """Union of two relations"""

    def __init__(self, left: Relation, right: Relation):
        """Unites left relation with right

        Parameters
        ----------
        left : Relation
            Left side of the operation
        right : Relation
            Right side of the operation

        Raises
        ------
        Exception
            If left and right have different schemas
        """
        if not left.get_schema() == right.get_schema():
            raise Exception("Cannot unite relations with different schemas")
        super().__init__(left.get_schema(), left, right)

    def can_be_infinite(self):
        return self.get_left().can_be_infinite() or self.get_right().can_be_infinite()

    def can_is_member_loop(self):
        return (
            self.get_left().can_is_member_loop()
            or self.get_right().can_is_member_loop()
        )

    def is_member(self, element):
        if self.get_left().can_is_member_loop():
            return self.get_right().is_member(element) or self.get_left().is_member(
                element
            )
        else:
            return self.get_left().is_member(element) or self.get_right().is_member(
                element
            )

    def members(self):
        """Yields tuples from left and right relations alternately"""
        return self._distinct(self._members)

    def _members(self):
        """Helper generator function that zips together tuples of both relations

        Doesn't handle duplicates
        """
        for item in zip_longest(self.get_left().members(), self.get_right().members()):
            if not item[0] == None:
                yield item[0]
            if not item[1] == None:
                yield item[1]



import unittest
from infinite_relations import *

# CONSTANTS
lr = list_relation(["fib"], [])
lr2 = list_relation(["fib"], [])
ir = fibonacci_relation()
ir2 = fibonacci_relation()


class BooleanFunctionsTest(unittest.TestCase):
    def test_trivial(self):
        self.assertFalse(lr.can_be_infinite())
        self.assertTrue(ir.can_be_infinite())

        self.assertFalse(lr.can_is_member_loop())
        self.assertTrue(ir.can_is_member_loop())

    def test_union(self):
        self.assertFalse(union(lr, lr2).can_be_infinite())
        self.assertTrue(union(lr, ir).can_be_infinite())
        self.assertTrue(union(ir, ir2).can_be_infinite())

        self.assertFalse(union(lr, lr2).can_is_member_loop())
        self.assertTrue(union(lr, ir).can_is_member_loop())
        self.assertTrue(union(ir, ir2).can_is_member_loop())

    def test_difference(self):
        self.assertFalse(difference(lr, lr2).can_be_infinite())
        self.assertFalse(difference(lr, ir).can_be_infinite())
        self.assertTrue(difference(ir, lr).can_be_infinite())
        self.assertTrue(difference(ir, ir2).can_be_infinite())

        self.assertFalse(difference(lr, lr2).can_is_member_loop())
        self.assertTrue(difference(lr, ir).can_is_member_loop())
        self.assertTrue(difference(ir, lr).can_is_member_loop())
        self.assertTrue(difference(ir, ir2).can_is_member_loop())

    def test_cross_join(self):
        self.assertFalse(cross_join(lr, lr2).can_be_infinite())
        self.assertTrue(cross_join(lr, ir).can_be_infinite())
        self.assertTrue(cross_join(ir, lr).can_be_infinite())
        self.assertTrue(cross_join(ir, ir2).can_be_infinite())

        self.assertFalse(cross_join(lr, lr2).can_is_member_loop())
        self.assertTrue(cross_join(lr, ir).can_is_member_loop())
        self.assertTrue(cross_join(ir, lr).can_is_member_loop())
        self.assertTrue(cross_join(ir, ir2).can_is_member_loop())

    def test_projection(self):
        self.assertFalse(projection(lr, ["fib"]).can_be_infinite())
        self.assertTrue(projection(ir, ["fib"]).can_be_infinite())

        self.assertFalse(projection(lr, ["fib"]).can_is_member_loop())
        self.assertTrue(projection(ir, ["fib"]).can_is_member_loop())

    def test_selection(self):
        self.assertFalse(selection(lr, empty_cond).can_be_infinite())
        self.assertTrue(selection(ir, empty_cond).can_be_infinite())

        self.assertFalse(selection(lr, empty_cond).can_is_member_loop())
        self.assertTrue(selection(ir, empty_cond).can_is_member_loop())

    def test_intersection(self):
        self.assertFalse(intersection(lr, lr2).can_be_infinite())
        self.assertFalse(intersection(lr, ir).can_be_infinite())
        self.assertTrue(intersection(ir, ir2).can_be_infinite())

        self.assertFalse(intersection(lr, lr2).can_is_member_loop())
        self.assertTrue(intersection(lr, ir).can_is_member_loop())
        self.assertTrue(intersection(ir, ir2).can_is_member_loop())

    def test_natural_join(self):
        self.assertFalse(natural_join(lr, lr2).can_be_infinite())
        self.assertTrue(natural_join(lr, ir).can_be_infinite())
        self.assertTrue(natural_join(ir, lr).can_be_infinite())
        self.assertTrue(natural_join(ir, ir2).can_be_infinite())

        self.assertFalse(natural_join(lr, lr2).can_is_member_loop())
        self.assertTrue(natural_join(lr, ir).can_is_member_loop())
        self.assertTrue(natural_join(ir, lr).can_is_member_loop())
        self.assertTrue(natural_join(ir, ir2).can_is_member_loop())

    def test_division(self):
        self.assertFalse(division(lr, lr2).can_be_infinite())
        self.assertFalse(division(lr, ir).can_be_infinite())
        self.assertTrue(division(ir, lr).can_be_infinite())
        self.assertTrue(division(ir, ir2).can_be_infinite())

        self.assertFalse(division(lr, lr2).can_is_member_loop())
        self.assertTrue(division(lr, ir).can_is_member_loop())
        self.assertTrue(division(ir, lr).can_is_member_loop())
        self.assertTrue(division(ir, ir2).can_is_member_loop())



"""Boolean functions to be used when creating selections
"""


def equality_cond(x: object, y: object) -> bool:
    return x == y


def add_cond(x: int, y: int, sum: int) -> bool:
    return x + y == sum


def empty_cond(*any) -> bool:
    return True



"""Utility functions for work with lists"""


def compare_lists_orderless(list: list, list_2: list) -> bool:
    """Returns True if both lists containg the same items"""
    return len(list) == len(list_2) and all(item in list_2 for item in list)


def flatten_list(list_of_lists: list) -> list:
    """Creates a list from a list of lists"""
    return [item for sublist in list_of_lists for item in sublist]


def dict_from_lists(keys: list, values: list) -> dict:
    """Creates a dict from a list of keys and list of values"""
    if len(values) != len(keys):
        raise Exception("Number of keys and values doesn't match")

    return dict(zip(keys, values))


def dict_values(dict: dict, keys: list) -> list:
    """Gets a list of values corresponding to a list of keys from a dict

    Parameters
    ----------
    dict : dict
        Dict to get values from
    keys : list
        Subset of keys of the dict

    Returns
    -------
    list
        List of dict's values corresponding to keys
    """
    return [dict[key] for key in keys]


def is_subset(list: list, list_2: list) -> bool:
    """Checks if list is a subset of list_2"""
    return all(item in list_2 for item in list)


def subtract(list: list, list_2: list) -> list:
    """Subtracts list_2 from list"""
    return [item for item in list if not item in list_2]



from __future__ import annotations
from abc import ABC, abstractclassmethod
from ..utils.relational_schema import RelationalSchema
from typing import Iterable


class Relation(ABC):
    """Abstract ancestor containing attributes and methods common to all relations

    Attributes
    ----------
    _schema : list
        A list of attribute names
    """

    def __init__(self, schema: list):
        """
        Parameters
        ----------
        schema : list
            A list of attribute names
        """
        self._schema: list = schema

    def get_schema(self) -> list:
        """Gets relation's schema"""
        return self._schema

    def rename(self, old_names: list, new_names: list) -> Relation:
        """Helper function to rename relation's attributes without having to create a new one

        Parameters
        ----------
        relation : Relation
            Relation to rename attributes of
        old_names : list
            List of attribute names to be changed
        new_names : list
            New attribute names

        Returns
        -------
        Relation
            Relation with a new schema
        """
        new_attrs = []
        for attr in self.get_schema():
            if attr in old_names:
                new_name = new_names[old_names.index(attr)]
                new_unique_name = RelationalSchema.get_unique_attr_name(
                    new_name, new_attrs
                )
                new_attrs.append(new_unique_name)
            else:
                new_attrs.append(attr)
        self._schema = new_attrs

    def can_be_infinite(self) -> bool:
        """
        Returns
        -------
        bool
            True, if the relation can contain an infinite number of tuples
        """
        return True

    def can_is_member_loop(self) -> bool:
        """
        Returns
        -------
        bool
            True, if is_member() can get stuck in an infinite loop without giving an answer
        """
        return True

    @abstractclassmethod
    def members(self) -> Iterable:
        """Iterable containing relation's tuples

        Every derived class must implement it's own members() method

        Returns
        -------
        Iterable
            Iterable containing relation's tuples
        """
        pass

    def list_members(self, count: int = -1) -> list:
        """Gets a given number of tuples in a list

        Good for use with infinite relations where you can't get all the members

        Parameters
        ----------
        count : int
            Number of tuples to extract, by default -1
            If not supplied, will yield all members of a list relation or first 10 members of an infinite relation

        Returns
        -------
        list
            List of first x member tuples
        """
        members = self.members()

        if count == -1:
            if not self.can_be_infinite():
                return list(members)
            else:
                count = 10

        member_list = []
        for _ in range(count):
            member_list.append(next(members))
        return member_list

    def is_member(self, tpl: dict) -> bool:
        """Checks if a tuple is contained in the relation

        Can potentially get stuck in an infinite loop without giving an answer.
        Check can_is_member_loop()

        Parameters
        ----------
        tpl : dict
            Tuple to check if it's contained in the relation

        Returns
        -------
        bool
            True, if tpl is contained in the relation
        """
        return tpl in self.members()

    def print(self, limit: int = 10_000) -> None:
        """Prints the schema then iterates over member tuples and prints their values in rows

        Parameters
        ----------
        limit : int
            Maximum number of printed members, useful for printing infinite relations
        """
        print("============")
        RelationalSchema.print(self.get_schema())
        print("------------")
        for tpl in self.members():
            print(*tpl.values())
            limit -= 1
            if limit == 0:
                break
        print("============")

    def _distinct(self, members: Iterable) -> Iterable:
        """Stores yielded tuples to avoid yielding duplicates

        Parameters
        ----------
        members: Iterable
            Iterable with potentially duplicate tuples

        Returns
        -------
        Iterable
            Iterable with no duplicates
        """
        yielded = []
        for item in members():
            if not item in yielded:
                yielded.append(item)
                yield item



from ..relations.relation import Relation
from ..utils.relational_schema import RelationalSchema
from ..utils.conditions import equality_cond
from ..utils.list_operations import is_subset, subtract
from .projection import Projection
from .difference import Difference
from .selection import Selection
from .cross_join import CrossJoin


def intersection(left: Relation, right: Relation) -> Relation:
    """Function representing intersection of two relations

    Parameters
    ----------
    left : Relation
        Left side of the operation
    right : Relation
        Right side of the operation

    Returns
    -------
    Relation
        Relation containing only tuples present in both input relations

    Raises
    ------
    Exception
        If schemas of input relations don't match
    """
    if not left.get_schema() == right.get_schema():
        raise Exception("Cannot intersect relations with different schemas")
    if left.can_be_infinite():
        return Difference(right, Difference(right, left))
    else:
        return Difference(left, Difference(left, right))


def natural_join(left: Relation, right: Relation) -> Relation:
    """Function representing a natural join of two relations

    Parameters
    ----------
    left : Relation
        Left side of the operation
    right : Relation
        Right side of the operation

    Returns
    -------
    Relation
        Relation containing all joined tuples of the input relations
    """
    join_schema, common_attrs = RelationalSchema.join(
        left.get_schema(), right.get_schema()
    )
    result = CrossJoin(left, right)

    for attr in common_attrs:
        result = Selection(result, equality_cond, [attr, attr + "'"])

    return Projection(result, join_schema)


def division(left: Relation, right: Relation) -> Relation:
    """Function representing the division of relations

    Parameters
    ----------
    left : Relation
        Relation to be divided
    right : Relation
        Divisor

    Returns
    -------
    Relation
        A relation that is the result of relational division

    Raises
    ------
    Exception
        If right relation's schema isn't a subset of left one's
    """
    l_schema = left.get_schema()
    r_schema = right.get_schema()
    if not is_subset(r_schema, l_schema):
        raise Exception(
            "Schema of a divisor relation must be a subset of the schema of the relation being divided."
        )

    l_schema_unique = subtract(l_schema, r_schema)
    l_unique = Projection(left, l_schema_unique)

    all_combinations = CrossJoin(l_unique, right)
    missing_combinations = Difference(all_combinations, left)

    return Difference(l_unique, Projection(missing_combinations, l_schema_unique))


def decomposition(r: Relation) -> list:
    """Function that should find functional dependencies in a relation and then split it
    into smaller ones until they all satisfy BCNF

    NOT IMPLEMENTED

    Parameters
    ----------
    r : Relation
        Relation to be decomposed

    Returns
    -------
    list
        A list of relations that all satisfy BCNF and when joined back equal the input relation
    """
    pass



from abc import ABC
from ..relations.relation import Relation


class UnaryOperation(Relation, ABC):
    """Abstract ancestor for any unary operation on a relation

    Attributes
    ----------
    _original_relation : Relation
        unmodified input of the operation
    """

    def __init__(self, relation: Relation):
        """
        Parameters
        ----------
        relation : Relation
            Input of the operation
        """
        super().__init__(relation.get_schema())
        self._original_relation: Relation = relation

    def get_original_relation(self) -> Relation:
        """Gets the unchanged input relation of the operation"""
        return self._original_relation

    def can_be_infinite(self):
        return self.get_original_relation().can_be_infinite()

    def can_is_member_loop(self):
        return self.get_original_relation().can_is_member_loop()



import unittest
from infinite_relations import *


# UTILS
def compare_relation_members(r1: Relation, r2: Relation) -> bool:
    """Compare first x tuples, or all"""
    return compare_lists_orderless(
        r1.get_schema(), r2.get_schema()
    ) and compare_lists_orderless(r1.list_members(), r2.list_members())


# CONSTANTS
schema = ["name", "age"]
names = ["Dani", "Emil", "Franta", "Martin", "Karel", "Standa", "Rudolf", "Tereza"]
ages = [23, 26, 25, 23, 28, 19, 25, 14]


# Test basic relational operations
class BasicOperationsTest(unittest.TestCase):
    def test_union(self):
        r1 = list_relation(
            schema,
            [{"name": name, "age": age} for (name, age) in zip(names[:5], ages[:5])],
        )
        r2 = list_relation(
            schema,
            [{"name": name, "age": age} for (name, age) in zip(names[3:8], ages[3:8])],
        )

        result = union(r1, r2)
        expected = list_relation(
            schema, [{"name": name, "age": age} for (name, age) in zip(names, ages)]
        )

        self.assertTrue(compare_relation_members(result, expected))

    def test_difference(self):
        r1 = list_relation(
            schema,
            [{"name": name, "age": age} for (name, age) in zip(names[:5], ages[:5])],
        )
        r2 = list_relation(
            schema,
            [{"name": name, "age": age} for (name, age) in zip(names[3:8], ages[3:8])],
        )

        result = difference(r1, r2)
        expected = list_relation(
            schema,
            [{"name": name, "age": age} for (name, age) in zip(names[:3], ages[:3])],
        )

        self.assertTrue(compare_relation_members(result, expected))

    def test_cross_join(self):
        r1 = list_relation(
            ["name", "age"],
            [{"name": name, "age": age} for (name, age) in zip(names[:3], ages[:3])],
        )
        r2 = list_relation(["round"], [{"round": 1}, {"round": 2}, {"round": 3}])

        result = cross_join(r1, r2)
        expected = list_relation(
            ["round", "name", "age"],
            [
                {"round": 1, "name": "Dani", "age": 23},
                {"round": 1, "name": "Emil", "age": 26},
                {"round": 1, "name": "Franta", "age": 25},
                {"round": 2, "name": "Dani", "age": 23},
                {"round": 2, "name": "Emil", "age": 26},
                {"round": 2, "name": "Franta", "age": 25},
                {"round": 3, "name": "Dani", "age": 23},
                {"round": 3, "name": "Emil", "age": 26},
                {"round": 3, "name": "Franta", "age": 25},
            ],
        )

        self.assertTrue(compare_relation_members(result, expected))

    def test_projection(self):
        r = list_relation(
            ["name", "age"],
            [{"name": name, "age": age} for (name, age) in zip(names, ages)],
        )

        result = projection(r, ["name"])
        expected = list_relation(["name"], [{"name": name} for name in names])

        self.assertTrue(compare_relation_members(result, expected))

    def test_selection(self):
        r = list_relation(
            ["name", "age"],
            [{"name": name, "age": age} for (name, age) in zip(names, ages)],
        )

        result = selection(r, lambda age: age > 25, ["age"])
        expected = list_relation(
            ["name", "age"],
            [{"name": "Emil", "age": 26}, {"name": "Karel", "age": 28}],
        )

        self.assertTrue(compare_relation_members(result, expected))


# Test operations derived from basic ones
class DerivedOperationsTest(unittest.TestCase):

    def test_intersection(self):
        r1 = list_relation(
            schema,
            [{"name": name, "age": age} for (name, age) in zip(names[:5], ages[:5])],
        )
        r2 = list_relation(
            schema,
            [{"name": name, "age": age} for (name, age) in zip(names[3:8], ages[3:8])],
        )

        result = intersection(r1, r2)
        expected = list_relation(
            schema, [{"name": "Martin", "age": 23}, {"name": "Karel", "age": 28}]
        )

        self.assertTrue(compare_relation_members(result, expected))

    def test_natural_join(self):
        r1 = list_relation(
            ["id"] + schema,
            [
                {"id": i + 1, "name": name, "age": age}
                for (i, (name, age)) in enumerate(zip(names, ages))
            ],
        )
        r2 = list_relation(
            ["id", "mark"], [{"id": 4, "mark": 99}, {"id": 6, "mark": 88}]
        )

        result = natural_join(r1, r2)
        expected = list_relation(
            ["id", "name", "age", "mark"],
            [
                {"id": 4, "name": "Martin", "age": 23, "mark": 99},
                {"id": 6, "name": "Standa", "age": 19, "mark": 88},
            ],
        )

        self.assertTrue(compare_relation_members(result, expected))

    def test_division(self):
        subjects = ["math", "math", "math", "math", "english", "english"]
        r1 = list_relation(
            ["name", "subject"],
            [
                {"name": name, "subject": sub}
                for (name, sub) in zip(names[:4] + names[:2], subjects)
            ],
        )
        r2 = list_relation(["subject"], [{"subject": "math"}, {"subject": "english"}])

        result = division(r1, r2)
        expected = list_relation(["name"], [{"name": "Dani"}, {"name": "Emil"}])

        self.assertTrue(compare_relation_members(result, expected))



from ..utils.list_operations import dict_from_lists
from ..utils.relational_schema import RelationalSchema
from .relation import Relation
from ..relations.list_relation import ListRelation
from ..operations.selection import Selection
from ..operations.cross_join import CrossJoin
from inspect import getfullargspec
import itertools
import re


class IntegerRelation(Relation):
    """Unary relation containing all non-negative integers"""

    def __init__(self, schema: list = ["n"]):
        """Creates unary integer relation with a given one-attribute schema

        Parameters
        ----------
        schema : list
            Relational schema with one attribute, by default ["n"]

        Raises
        ------
        Exception
            If schema doesn't have exactly 1 attribute
        """
        if len(schema) == 1:
            super().__init__(schema)
        else:
            raise Exception("Schema of IntegerRelation must have 1 attribute")

    def can_be_infinite(self):
        return True

    def can_is_member_loop(self):
        return False

    def members(self):
        """Yields an infinite sequence of non-negative integers in order"""
        i = 0
        while True:
            yield dict_from_lists(self.get_schema(), [i])
            i += 1

    def is_member(self, element: dict):
        vals = list(element.values())
        if len(vals) == 1:
            val = vals[0]
            return isinstance(val, int) and val >= 0
        return False


class GeneralIntegerRelation(Selection):
    """Infinite relation containing integer tuples which satisfy a given condition"""

    def __init__(self, schema: list, condition: callable):
        """Creates an infinite relation of integer tuples determined by a given schema and condition

        Parameters
        ----------
        schema : list
            A list of attribute names
        condition : callable
            Condition that tuples must satisfy

        Raises
        ------
        Exception
            If length of schema doesn't match number of condition parameters
        """
        len_cond_params = len(getfullargspec(condition).args)
        if len_cond_params != len(schema) and len_cond_params != 0:
            raise Exception("Schema length must match number of condition parameters")

        if len(schema) == 0:
            result = ListRelation([])
        else:
            result = IntegerRelation([schema[0]])
            for attr in schema[1:]:
                result = CrossJoin(result, IntegerRelation([attr]))
        super().__init__(result, condition, schema)

    def can_be_infinite(self):
        return True

    def can_is_member_loop(self):
        return False


class AddRelation(Relation):
    """Ternary relation containing all possible sums of non-negative integers"""

    def __init__(self, schema):
        """Creates an add relation with a given schema

        Parameters
        ----------
        schema : RelationalSchema, optional
            Relational schema

        Raises
        ------
        Exception
            If length of schema is not 3
        """
        if len(schema) == 3:
            super().__init__(schema)
        else:
            raise Exception("Schema of AddRelation must have 3 attributes")

    def can_be_infinite(self):
        return True

    def can_is_member_loop(self):
        return False

    def members(self):
        """Yields all possible triplets of numbers where x + y == z"""
        i = 0
        while True:
            for j in range(i + 1):
                yield dict_from_lists(self.get_schema(), [j, i - j, i])
            i += 1

    def is_member(self, tpl: dict) -> bool:
        vals = list(tpl.values())
        if not len(vals) == 3:
            return False
        for val in vals:
            if not isinstance(val, int):
                return False
        return vals[0] + vals[1] == vals[2]


class FibonacciRelation(Relation):
    """Unary relation containing all members of the fibonacci sequence"""

    def __init__(self):
        """Creates unary relation containing the fibonacci sequence"""
        super().__init__(["fib"])

    def members(self):
        """Generates the fibonacci sequence"""
        f1 = 0
        f2 = 1
        while True:
            yield dict_from_lists(self.get_schema(), [f1])
            f1, f2 = f2, f1 + f2


class EqualRelation(Relation):
    """Binary integer relation where every tuple consists of two equal numbers"""

    def __init__(self, schema: list = ["x", "y"]):
        """Creates binary relation of equal numbers"""
        if not len(schema) == 2:
            raise Exception("Schema of EqualRelation must have 2 attributes")
        super().__init__(schema)

    def can_is_member_loop(self) -> bool:
        return False

    def is_member(self, tpl: dict) -> bool:
        vals = list(tpl.values())
        if not len(vals) == 2:
            return False
        for val in vals:
            if not isinstance(val, int):
                return False
        return vals[0] == vals[1]

    def members(self):
        """Yields tuples of two equal integers"""
        i = 0
        while True:
            yield dict_from_lists(self.get_schema(), [i, i])
            i += 1


class UnequalRelation(Relation):
    """Binary integer relation where every tuple consists of two unequal numbers"""

    def __init__(self):
        """Creates binary relation of unequal numbers"""
        super().__init__(["lesser", "greater"])

    def can_is_member_loop(self) -> bool:
        return False

    def is_member(self, tpl: dict) -> bool:
        vals = list(tpl.values())
        if not len(vals) == 2:
            return False
        for val in vals:
            if not isinstance(val, int):
                return False
        return vals[0] < vals[1]

    def members(self):
        """Yields tuples of two equal integers"""
        i = 0
        while True:
            for j in range(i):
                yield dict_from_lists(self.get_schema(), [j, i])
            i += 1


class WordRelation(Relation):
    """Relation of all possible combinations of the standard 26 letters of the alphabet (lowercase),
    starting from one-letter words and getting longer.
    """

    alphabet = [
        "a",
        "b",
        "c",
        "d",
        "e",
        "f",
        "g",
        "h",
        "i",
        "j",
        "k",
        "l",
        "m",
        "n",
        "o",
        "p",
        "q",
        "r",
        "s",
        "t",
        "u",
        "v",
        "w",
        "x",
        "y",
        "z",
    ]

    def __init__(self, schema: list = ["word"]):
        """Creates relation of all possible words"""
        if not len(schema) == 1:
            raise Exception("Schema of WordRelation must have 1 attributes")
        super().__init__(schema)

    def can_is_member_loop(self) -> bool:
        return False

    def is_member(self, tpl: dict) -> bool:
        return len(tpl) == 1 and re.fullmatch(tpl[self.get_schema()[0]], r"[a-z]+")

    def members(self):
        """Yields all possible words"""
        i = 0
        while True:
            for combination in itertools.product(self.alphabet, repeat=i):
                yield dict_from_lists(self.get_schema(), ["".join(combination)])
            i += 1



from __future__ import annotations
from .relation import Relation


class ListRelation(Relation):
    """Relation with a finite number of members, initiated by a list

    Attributes
    ----------
    _schema : list
        A list of attributes
    _members : list
        A list of contained tuples
    """

    def __init__(self, schema: list, tuples: list = []):
        """Creates a finite relation from a given schema and a list of tuples

        Parameters
        ----------
        schema : list
            A list of attribute names
        tuples : list, optional
            A list of tuples (dicts) to initialize the relation, by default []

        Raises
        ------
        Exception
            If length of schema and a tuple don't match
        TypeError
            If a tuple is not a dict
        """
        for tuple in tuples:
            if len(schema) != len(tuple):
                raise Exception("Length of schema and tuples must match")
            if not isinstance(tuple, dict):
                raise TypeError("Tuples must be dictionaries")
        super().__init__(schema)
        self._members = tuples

    def can_be_infinite(self):
        return False

    def can_is_member_loop(self):
        return False

    def members(self):
        """Returns an iterator from a finite member list"""
        return iter(self._members)



from ..relations.relation import Relation
from .binary_operation import BinaryOperation


class Difference(BinaryOperation):
    """Difference between two relations"""

    def __init__(self, left: Relation, right: Relation):
        """Subtracts right relation from left

        Not commutative

        Parameters
        ----------
        left : Relation
            Left side of the operation
        right : Relation
            Right side of the operation

        Raises
        ------
        Exception
            If left and right have different schemas
        """
        if not left.get_schema() == right.get_schema():
            raise Exception("Cannot subtract relations with different schemas")
        super().__init__(left.get_schema(), left, right)

    def can_be_infinite(self):
        return self.get_left().can_be_infinite()

    def can_is_member_loop(self):
        return (
            self.get_left().can_is_member_loop()
            or self.get_right().can_is_member_loop()
        )

    def is_member(self, element):
        return self.get_left().is_member(element) and not self.get_right().is_member(
            element
        )

    def members(self):
        """Yields all tuples of left relation that are absent in the right"""
        return (
            tpl
            for tpl in self.get_left().members()
            if not self.get_right().is_member(tpl)
        )



from __future__ import annotations

from ..relations.relation import Relation
from ..utils.list_operations import is_subset
from .unary_operation import UnaryOperation


class Projection(UnaryOperation):
    """Relation projected onto a new schema"""

    def __init__(self, relation: Relation, new_schema: list):
        """Projects relation onto new schema

        Parameters
        ----------
        relation : Relation
            Input relation of the operation
        new_schema : list
            New schema to project the relation onto

        Raises
        ------
            Exception
                If new schema isn't a subset of the relation's current schema
        """
        if is_subset(new_schema, relation.get_schema()):
            super().__init__(relation)
            self._new_schema = new_schema
        else:
            raise Exception("Schema of projection must be a subset of original schema")

    def can_is_member_loop(self):
        return self.can_be_infinite()

    def members(self):
        """Yields all tuples of the original relation projected on the new schema"""
        return self._distinct(self._members)

    def _members(self):
        """Helper generator function that yields projected tuples

        Doesn't handle duplicates
        """
        for tpl in self.get_original_relation().members():
            yield {attr: tpl[attr] for attr in self.get_schema()}

    def get_schema(self):
        return self._new_schema



from __future__ import annotations

from ..utils.list_operations import dict_values
from ..relations.relation import Relation
from .unary_operation import UnaryOperation
from inspect import getfullargspec


class Selection(UnaryOperation):
    """Class representing a selection (restriction) of a relation"""

    def __init__(
        self, relation: Relation, condition: function, condition_param_names: list
    ):
        """Restricts the relation to tuples that satisfy a condition

        Parameters
        ----------
        relation : Relation
            Input relation of the operation
        condition : function
            Condition, that tuples must satisfy
        condition_param_names : list
            List of attribute names to be evaluated by the condition

        Raises
        ------
        Exception
            If number of condition parameters isn't empty and doesn't match the condition
        """
        arglen = len(getfullargspec(condition).args)
        if arglen == 0 or arglen == len(condition_param_names):
            self.condition_param_names = condition_param_names
        elif arglen == len(relation.get_schema()):
            self.condition_param_names = relation.get_schema()
        else:
            raise Exception(
                "Number of parameters in condition must either a) be 0; b) match condition_param_names length; c) match relation schema length."
            )
        super().__init__(relation)
        self.condition = condition

    def is_member(self, tpl):
        return self.condition(
            *dict_values(tpl, self.condition_param_names)
        ) and super().is_member(tpl)

    def members(self):
        """Yields tuples from the original relation if they satisfy the condition"""
        for tpl in self.get_original_relation().members():
            condition_params = dict_values(tpl, self.condition_param_names)
            if self.condition(*condition_params):
                yield tpl



import unittest
from infinite_relations import *


# UTILS
def compare_relation_members(r1: Relation, r2: Relation, c: int) -> bool:
    """Compare first x tuples, or all"""
    return compare_lists_orderless(
        r1.get_schema(), r2.get_schema()
    ) and compare_lists_orderless(r1.list_members(c), r2.list_members(c))


# CONSTANTS
naturals = integer_relation()
odds = general_integer_relation(["n"], lambda n: n % 2 == 1)
evens = general_integer_relation(["n"], lambda n: n % 2 == 0)

add_schema = ["x", "y", "sum"]
raw_add = general_integer_relation(add_schema, add_cond)


# Test basic relational operations
class BasicOperationsTest(unittest.TestCase):
    def test_union(self):
        result = union(evens, odds)
        expected = naturals

        self.assertTrue(compare_relation_members(result, expected, 20))

    def test_difference(self):
        result = difference(naturals, odds)
        expected = evens

        self.assertTrue(compare_relation_members(result, expected, 20))

    def test_cross_join(self):
        r1 = list_relation(
            ["name", "age"],
            [
                {"name": "Dani", "age": 23},
                {"name": "Emil", "age": 26},
                {"name": "Franta", "age": 25},
            ],
        )
        r2 = general_integer_relation(["round"], lambda x: x > 0)

        result = cross_join(r1, r2)
        expected = list_relation(
            ["round", "name", "age"],
            [
                {"round": 1, "name": "Dani", "age": 23},
                {"round": 1, "name": "Emil", "age": 26},
                {"round": 1, "name": "Franta", "age": 25},
                {"round": 2, "name": "Dani", "age": 23},
                {"round": 2, "name": "Emil", "age": 26},
                {"round": 2, "name": "Franta", "age": 25},
                {"round": 3, "name": "Dani", "age": 23},
                {"round": 3, "name": "Emil", "age": 26},
                {"round": 3, "name": "Franta", "age": 25},
                {"round": 4, "name": "Dani", "age": 23},
            ],
        )

        self.assertTrue(compare_relation_members(result, expected, 10))

    def test_projection(self):
        r = general_integer_relation(["x", "y", "z"], empty_cond)

        result = projection(r, ["x", "y"])
        expected = general_integer_relation(["x", "y"], empty_cond)

        self.assertTrue(compare_relation_members(result, expected, 20))

    def test_selection(self):
        result = selection(naturals, lambda n: n < 10)
        expected = list_relation(
            ["n"],
            [
                {"n": 0},
                {"n": 1},
                {"n": 2},
                {"n": 3},
                {"n": 4},
                {"n": 5},
                {"n": 6},
                {"n": 7},
                {"n": 8},
                {"n": 9},
            ],
        )

        self.assertTrue(compare_relation_members(result, expected, 10))


# Test operations derived from basic ones
class DerivedOperationsTest(unittest.TestCase):
    def test_intersection(self):
        result = intersection(naturals, evens)

        self.assertTrue(compare_relation_members(result, evens, 20))

    def test_natural_join(self):
        r1 = list_relation(
            ["name", "english_mark", "math_mark"],
            [
                {"name": "Dani", "english_mark": 90, "math_mark": 85},
                {"name": "Emil", "english_mark": 52, "math_mark": 99},
                {"name": "Franta", "english_mark": 70, "math_mark": 65},
            ],
        )

        result = natural_join(
            r1, add_relation(["english_mark", "math_mark", "mark_sum"])
        )
        expected = list_relation(
            ["name", "english_mark", "math_mark", "mark_sum"],
            [
                {"name": "Dani", "english_mark": 90, "math_mark": 85, "mark_sum": 175},
                {"name": "Emil", "english_mark": 52, "math_mark": 99, "mark_sum": 151},
                {
                    "name": "Franta",
                    "english_mark": 70,
                    "math_mark": 65,
                    "mark_sum": 135,
                },
            ],
        )

        self.assertTrue(compare_relation_members(result, expected, 3))

    def test_division(self):
        return
        r1 = general_integer_relation(["num"], empty_cond)
        r2 = list_relation(["char"], [{"char": "a"}, {"char": "b"}, {"char": "c"}])
        r3 = cross_join(r1, r2)
        r4 = list_relation(
            ["num", "char"], [{"num": 2, "char": "b"}, {"num": 3, "char": "c"}]
        )
        r5 = difference(r3, r4)

        result = division(r5, r2)
        expected = list_relation(
            ["num"],
            [
                {"num": 1},
                {"num": 4},
                {"num": 5},
                {"num": 6},
                {"num": 7},
            ],
        )

        self.assertTrue(compare_relation_members(result, expected, 5))
